import torch # PyTorchライブラリの読み込み
import torch.nn as nn # ニューラルネットワークライブラリをnnとして参照
import torch.nn.functional as F # 出力関数ライブラリをFとして参照
import torchvision # コンピュータビジョン関係ライブラリtorchvisionの読み込み
import torchvision.datasets as dsets # データセットをdsetsとして参照
import torchvision.transforms as transforms # データ変換関数の読み込み
import torch.optim # 最適化ライブラリ
import torch.utils.data # データ読み込みライブラリの読み込み

class MyPerceptron(nn.Module): #多層ネットワークの構造定義
    def __init__(self, input_size, hidden_size, num_classes):
        super(MyPerceptron, self).__init__() #親クラスのコンストラクタを呼び出す
        self.fc1 = nn.Linear(input_size, hidden_size) #1層目の定義
        self.fc2 = nn.Linear(hidden_size, hidden_size) #2層目の定義
        self.fc3 = nn.Linear(hidden_size, num_classes) #3層目の定義
        
    def forward(self, x):     #出力計算メソッド定義
        out = self.fc1(x)     #1層目のnet値計算
        out = F.relu(out)     #1層目の活性化関数
        out = self.fc2(out)   #2層目のnet値計算
        out = F.relu(out)     #2層目の活性化関数
        out = self.fc3(out)   #3層目のnet値計算
        return out #3層目の活性化関数を線形関数とする
     
class main():
    def __init__(self):
        self.input_size = 784
        self.hidden_size = 500
        self.num_classes = 10
        self.batch_size = 100
        self.train_dataset = dsets.MNIST(root='./data', train=True, transform=transforms.ToTensor(),download=True) # 学習データセット読み込み
        train_loader = torch.utils.data.DataLoader(dataset=self.train_dataset, batch_size=self.batch_size, shuffle=True, pin_memory=False) # 学習データをバッチサイズに分割する。
        # GPUもしくはCPUの選択　deviceの自動選択 #
        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        #self.device = 'cpu'
        print("device = %s" %(self.device))
        self.criterion = nn.CrossEntropyLoss() # ClossEntropyロスを評価関数とする
        # パーセプトロンクラスのインスタンス化 #
        self.model = MyPerceptron(self.input_size, self.hidden_size, self.num_classes)
        # GPUもしくはCPUのどちらで処理をするのかをパーセプトロンのオブジェクトに指定 #
        self.model = self.model.to(self.device)
        # 最適化理論としてAdagradを採用 #
        self.optimizer = torch.optim.Adagrad(self.model.parameters(True), lr=0.01)

        for i in range(100): #100エポック繰り返す
            totalloss = 0
            totalacc = 0
            for index, (images, labels) in enumerate(train_loader): 
                self.optimizer.zero_grad()       # 勾配情報を0に初期化
                images = images.view(-1,28*28)   # 28x28のサイズで学習
                images = images.to(self.device).detach()
                labels = labels.to(self.device).detach()
                outputs = self.model.forward(images)   # forward計算し最終出力を求める
                acc = (self.getLabel(outputs) == labels).sum() # 正解数を数える
                totalacc += acc.item()
                
                loss = self.criterion(outputs, labels) # lossを計算する
                totalloss += loss.item()         # totallossにlossを加える
                loss.backward()                  # 勾配計算(back propagation)
                self.optimizer.step()            # パラメータを更新

            print("%s : err=%s acc=%s" %(i, totalloss, totalacc))
        train_loss = totalloss / len(train_loader)
        print("averaged train loss = %s" %(train_loss))
        
    def getLabel(self, outputs):
        return outputs.max(1)[1]
        
if __name__ == "__main__":
    MAIN = main()
