import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data.dataset import Subset
import torchvision.datasets as dsets
import torchvision.transforms as transforms
import torch.optim
import torch.utils.data
import copy

class MyPerceptron(nn.Module): #多層ネットワークの構造定義
    def __init__(self, input_size, hidden_size, num_classes,device): #input_size:入力サイズ, hidden_size:中間ユニット数, num_classes:出力ユニット数, device:CPU/GPU
        super(MyPerceptron, self).__init__()
        self.device = device
        self.fc1 = nn.Linear(input_size, hidden_size).to(self.device)
        self.fc2 = nn.Linear(hidden_size, hidden_size).to(self.device)
        self.fc3 = nn.Linear(hidden_size, num_classes).to(self.device)
        self.criterion = nn.CrossEntropyLoss().to(self.device) #nn.CrossEntropyLoss().to(device)
        self.num_classes = num_classes
        self.optimizer = torch.optim.Adagrad(self.parameters(True), lr=0.01)
        
    def forward(self, x): #出力計算メソッド定義
        x = x.to(self.device)
        out = self.fc1(x)
        out = F.relu(out)
        out = self.fc2(out)
        out = F.relu(out)
        out = self.fc3(out)
        return F.softmax(out, dim=0)
    
    def learning(self, train_loader):
        self.train() # 学習モード
        totalloss = 0 # 変数初期化
        totalacc = 0 # 変数初期化2
        for index, (images, labels) in enumerate(train_loader):
            images = images.view(-1,28*28)
            if index == 0:
                print("MyPerceptron.learning() labels=%s" %(labels))
            self.optimizer.zero_grad()
            outputs = self(images)
            outputs = outputs.to(self.device) #
            labels = labels.to(self.device)
            loss = self.criterion(outputs, labels)
            totalloss += loss.item()
            acc = (self.getLabel(outputs) == labels).sum() # 正解数を数える
            totalacc += acc.item()
            loss.backward()
            self.optimizer.step()
        return totalloss, totalacc   

    def evaluation(self, data_loader):
        totalacc = 0
        each_acc = torch.zeros([10,10])
        class_size = torch.zeros([10])
        for index, (images, labels) in enumerate(data_loader):
            images = images.view(-1, 28*28)
            outputs = self(images)
            outputs = outputs.to(self.device)
            labels = labels.to(self.device)
            acc = (self.getLabel(outputs)==labels).sum()
            actuallabels = self.getLabel(outputs)
            for p in range(len(labels)):
                    each_acc[labels[p].item()][actuallabels[p].item()] += 1
                    class_size[labels[p].item()] += 1
            totalacc += acc.item()
        return totalacc, each_acc, class_size
         
    def getLabel(self, outputs):
#         print("getLabel() outputs=%s label=%s" %(outputs, outputs.max(1)[1]) )
        return outputs.max(1)[1]
  
class MyMNIST(torch.utils.data.Dataset):
    # MinistデータセットのうちTargetLabels[]で指定された特定のラベルのデータだけを集める
    def __init__(self, TargetLabels, root, train, transform, download):
        self.transform = transform
        self.download = download
        self.tran = train
        self.dataset = dsets.MNIST(root='./data', train=True, transform=self.transform, download=self.download)
        self.datasetSize = 0
        self.indicies = []
        self.label = []
        for p in range(self.dataset.__len__()):
            each_input, each_label = self.dataset.__getitem__(p)
            for n in range(len(TargetLabels)):
                if each_label == TargetLabels[n]:#ラベルが一致する場合
                    self.indicies.append(p) # インデックスをself.indiciesに保持する
                    self.label.append(each_label) #ラベルデータもself.labelに保持する
                    self.datasetSize += 1 #ラベルの一致するデータの個数を数える
        self.mydata = torch.tensor([self.datasetSize,28,28]) #テンソル型の入力データ配列を用意する
        self.mydata = torch.zeros(self.datasetSize, 28, 28) #上で用意したテンソル型データを初期化
        for p in range(self.datasetSize): #ラベルの一致する入力データセットを作成する
            each_input, each_label = self.dataset.__getitem__(self.indicies[p])
            self.mydata[p] = each_input
        self.data = self.mydata.numpy() # 入力データはTensorからnumpy arrayに変換する必要がある
        
    def __len__(self): #データセットのサイズを返すメソッド
        return self.datasetSize
    
    def __getitem__(self, idx): #各々のデータを返すメソッド
        out_data = self.data[idx]
        out_label =  self.label[idx]
        if self.transform:
            out_data = self.transform(out_data) #ラベルデータの変換
        return out_data, out_label 
    
class main():     
    def __init__(self, num_division):
        print(" main() _init_() started!")
        self.input_size = 784
        self.hidden_size = 500
        self.num_classes = 10
        self.batch_size = 50
        self.num_division = num_division
        
        self.label_step = int(10/self.num_division)

        # テストデータセットを先に用意        
        self.test_dataset = dsets.MNIST(root='./data', train=False, transform=transforms.ToTensor(), download=True) #<--訂正：download=Trueを追加
        test_loader = torch.utils.data.DataLoader(dataset=self.test_dataset, 
                                                  batch_size=self.batch_size, shuffle=False)

        # GPUもしくはCPUの選択　deviceの自動選択
        device = 'cuda' if torch.cuda.is_available() else 'cpu'

        # パーセプトロンクラスのインスタンス化
        self.model = MyPerceptron(self.input_size, self.hidden_size, self.num_classes,device) # パーセプトロンのインスタンス化
        self.model = self.model.to(device)
        
        Acc = []
        for n in range(0, self.num_division):
            target_labels = []
            for l in range(n*self.label_step, (n+1)*self.label_step):
                target_labels.append(l)
                
               # 学習用・テスト用データをロードする
            self.train_dataset = MyMNIST(target_labels, './data', True, transforms.ToTensor(),True)
            totalloss = 0 #変数初期化
            totalacc = 0  #変数初期化
            for i in range(100): # 100回繰り返す
                train_loader = torch.utils.data.DataLoader(dataset=self.train_dataset, batch_size=100, shuffle=True)
                err, acc = self.model.learning(train_loader) # ミニバッチ学習
                print("%s : err=%s acc=%s" %(i, err, acc))
                totalloss += err
                totalacc += acc
                train_loss = totalloss / (100*len(self.train_dataset)) # 平均誤差算出
                totalacc = totalacc / (100*len(self.train_dataset)) # 平均正解率算出
            print("averaged train loss = %s, accuracy = %s" %(train_loss, totalacc)) #<--訂正：インデント修正
            testerr, AccMatrix, class_size = self.model.evaluation(test_loader) #テスト用サンプルでの誤差計測
            print("Accuracy matrix = %s" %(AccMatrix))
            for c in range(10):
                print("%s : %s " %(c, (float)(AccMatrix[c][c].item()/class_size[c].item())))
            Acc.append(testerr)
        #誤差の一覧表示
        for n in range(0, len(Acc)):
            print("Acc[%s]=%s" %(n, Acc[n]))
            
if __name__ == "__main__":
    MAIN =  main(5)
