import sys
import torch.nn as nn
import torch.nn.functional as F
#from torch.utils.data.dataset import Subset
import torchvision.datasets as dsets
import torchvision.transforms as transforms
import torch.optim
import torch.utils.data
# import pandas as pd # for reading CSV
import random


class MyPerceptron(nn.Module): #多層ネットワークの構造定義
    def __init__(self, input_size, hidden_size, num_classes,device): #input_size:入力サイズ, hidden_size:中間ユニット数, num_classes:出力ユニット数, device:CPU/GPU
        super(MyPerceptron, self).__init__()
        self.device = device
        self.fc1 = nn.Linear(input_size, hidden_size).to(self.device)
        self.fc2 = nn.Linear(hidden_size, hidden_size).to(self.device)
        self.fc3 = nn.Linear(hidden_size, num_classes).to(self.device)
        self.num_classes = num_classes
        
        
    def forward(self, x): #出力計算メソッド定義
        x = x.to(self.device)
        out = self.fc1(x)
        out = F.relu(out)
        out = self.fc2(out)
        out = F.relu(out)
        out = self.fc3(out)
        return out
         

 
class SBuffer(torch.utils.data.Dataset): #SBufferクラスは小容量のバッファを表現する Datasetクラスを実装
    def __init__(self, maxbuffersize, inputXsize, inputYsize): #<--訂正：引数labelSizeは不使用のため削除
        self.maxbuffersize = maxbuffersize
        self.inputXsize = inputXsize
        self.inputYsize = inputYsize
        #self.labelSize = labelSize <--訂正：以降使用していないためコメントアウト
        self.InputSTensor = torch.zeros(self.maxbuffersize, self.inputXsize, self.inputYsize)
        self.InputTensor = self.InputSTensor.numpy()
        self.currentBufferSize = 0
        self.Labels = []
        

    def __len__(self): #データセットのサイズを返すメソッド
        return self.currentBufferSize
    
    def __getitem__(self, idx): #各々のデータを返すメソッド
        out_data = self.InputTensor[idx]
        out_label =  self.Labels[idx]
        return out_data, out_label
        
    def pushBuffer(self, input, output): #一サンプルをバッファにセーブする。このときバッファサイズを一定に保つためにランダムに置き換え対象となるレコードを指定する
        
        if self.currentBufferSize < self.maxbuffersize:
            self.InputSTensor[self.currentBufferSize] = input.clone()
            self.Labels.append(output)
            self.currentBufferSize += 1
        else:
            targetIndex = random.randint(0,self.maxbuffersize-1)
#             print("SBuffer.pushBuffer() targetIndex=%s" %(targetIndex))
            self.InputSTensor[targetIndex] = input.clone()
            self.Labels[targetIndex] = output         
        self.InputTensor = self.InputSTensor.clone()
        

    
class MyMNISTWithBuffer(torch.utils.data.Dataset):
    # Ministデータセットとバッファに貯めてある古いデータの両方を取り出す
    def __init__(self, TargetLabels, root, train, transform, download, sbuffer):
          #引数で得られたデータを内部変数に保持する
        self.transform = transform
        self.download = download
        self.tran = train
        #MNISTおよびバッファデータを取り出すクラスをインスタンス化
        self.newDataset = dsets.MNIST(root='./data', train=True, transform=self.transform, download=self.download)
        self.oldDataset = sbuffer
          #変数の初期化
        self.datasetSize = 0
        self.indicies = []
        self.label = []
        
          #ラベルの一致するデータの個数を数える
        for p in range(self.newDataset.__len__()):
            each_input, each_label = self.newDataset.__getitem__(p)
            for n in range(len(TargetLabels)):
                if each_label == TargetLabels[n]:#ラベルが一致する場合
                    self.indicies.append(p) # インデックスをself.indiciesに保持する
                    self.datasetSize += 1 #ラベルの一致するデータの個数を数える
        print("MyMNISTWithBuffer() newdatasetSize=%s" %(self.datasetSize))
          #配列の割付           
        self.datasetSize += self.oldDataset.__len__() # データのサイズは新しいデータと古いデータのサイズを合計したもの
        print("MyMNISTWithBuffer() olddatasetSize=%s" %(self.oldDataset.__len__()))
        print("MyMNISTWithBuffer() totaldatasetSize=%s" %(self.datasetSize))
        self.mydata = torch.tensor([self.datasetSize,28,28]) #テンソル型の入力データ配列を用意する
        self.mydata = torch.zeros(self.datasetSize, 28, 28) #上で用意したテンソル型データを初期化
          #新規サンプルの取り出し
        for p in range(len(self.indicies)): #ラベルの一致する入力データセットを作成する
            each_input, each_label = self.newDataset.__getitem__(self.indicies[p])
            self.mydata[p] = each_input
            self.label.append(each_label) #ラベルデータもself.labelに保持する
#         print("MyMNISTWithBuffer() new sample label=%s" %(self.label))
#         古いサンプルの取り出し
        for p in range(self.oldDataset.__len__()):#古いデータを取り出す
            each_input, each_label = self.oldDataset.__getitem__(p)
            self.mydata[p+len(self.indicies)]=each_input # 新しいデータの後ろに古いを入れる
#             print("MyMNISTWithBuffer() each old sample label=%s" %(each_label))
            self.label.append(each_label) #ラベルもここで取り出す
        self.data = self.mydata.numpy() # 入力データはTensorからnumpy arrayに変換する必要がある
        
    def __len__(self): #データセットのサイズを返すメソッド
        return self.datasetSize
    
    def __getitem__(self, idx): #各々のデータを返すメソッド
        out_data = self.data[idx]
        out_label =  self.label[idx]
        if self.transform:
            out_data = self.transform(out_data) #ラベルデータの変換
        return out_data, out_label
     
    def pushBuffer(self):
        print("MyMNISTWithBuffer.pushBuffer() --storing the learned samples to the buffer. Please wait for a while.") #<--追加
        for p in range(len(self.indicies)): #ラベルの一致する入力データセットを作成する
            each_input, each_label = self.newDataset.__getitem__(self.indicies[p])
            print(".",end="") #<--追加
#             print("MyMINISTWithBuffer.pushBuffer() pushed label is %s" %(each_label)) #<--何が保存されたのかをチェックしたい場合にはコメントを外して下さい。
            self.oldDataset.pushBuffer(each_input, each_label)
            
class main():     
    def __init__(self, num_division, bufferSizeGain):
        print(" main() _init_() started!")
        self.input_size = 784
        self.hidden_size = 500
        self.num_classes = 10
        self.batch_size = 50
        self.num_division = num_division

        
        self.label_step = int(10/self.num_division)

          # テストデータセットを先に用意        
        self.test_dataset = dsets.MNIST(root='./data', train=False, transform=transforms.ToTensor(), download=True)
        test_loader = torch.utils.data.DataLoader(dataset=self.test_dataset, 
                                                  batch_size=self.batch_size, shuffle=False)

        # GPUもしくはCPUの選択　deviceの自動選択
        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
#         device = 'cpu' # GPUの場合には 'cuda:0'
        self.criterion = nn.CrossEntropyLoss().to(self.device) #nn.CrossEntropyLoss().to(device)
        self.model = MyPerceptron(self.input_size, self.hidden_size, self.num_classes, self.device) # パーセプトロンのインスタンス化
        self.model = self.model.to(self.device) # 計算するデバイスに適応させる      
        self.optimizer = torch.optim.Adagrad(self.model.parameters(True), lr=0.01)
        sampleSizePerRound = int(60000/self.num_division)
        buffersize = int(sampleSizePerRound * bufferSizeGain)
        self.sbuffer = SBuffer(buffersize,28,28) #訂正：第四引数は廃止したため第四引数に渡していた10を削除
        print("buffersize = %s (sample size per round: %s) bufferSizeGain=%s" %(buffersize, sampleSizePerRound, bufferSizeGain))
        Acc = []
        for n in range(0, self.num_division):
            target_labels = []
            for l in range(n*self.label_step, (n+1)*self.label_step):
                target_labels.append(l)
                
               # 学習用・テスト用データをロードする
            self.train_dataset = MyMNISTWithBuffer(target_labels, './data', True, transforms.ToTensor(),True,self.sbuffer)
            totalloss = 0 #変数初期化
            totalacc = 0  #変数初期化
            for i in range(100): # 100回繰り返す
                train_loader = torch.utils.data.DataLoader(dataset=self.train_dataset, batch_size=100, shuffle=True)
                err, acc = self.learning(self.model, train_loader) # ミニバッチ学習
                print("%s : err=%s acc=%s" %(i, err, acc))
                totalloss += err
                totalacc += acc
                train_loss = totalloss / (100*len(self.train_dataset)) # 平均誤差算出
                totalacc = totalacc / (100*len(self.train_dataset)) # 平均正解率算出
                print("averaged train loss = %s, accuracy = %s" %(train_loss, totalacc)) #　表示
            testerr, AccMatrix, class_size = self.evaluation(test_loader) #テスト用サンプルでの誤差計測
            print("Accuracy matrix = %s" %(AccMatrix))
            for c in range(10):
                print("%s : %s " %(c, (float)(AccMatrix[c][c].item()/class_size[c].item())))
            Acc.append(testerr)
            self.train_dataset.pushBuffer()
          #誤差の一覧表示
        for n in range(0, len(Acc)):
            print("Acc[%s]=%s" %(n, Acc[n]))
            
    def learning(self, model, train_loader):
        #self.train() # 学習モード
        totalloss = 0 # 変数初期化
        totalacc = 0 # 変数初期化2
        for index, (images, labels) in enumerate(train_loader):
            images = images.view(-1,28*28)
            if index==0:
                print("Main.learning() labels=%s" %(labels))
            self.optimizer.zero_grad()
            outputs = model(images)
            outputs = outputs.to(self.device) #
            labels = labels.to(self.device)
            loss = self.criterion(outputs, labels)
            totalloss += loss.item()
            acc = (self.getLabel(outputs) == labels).sum() # 正解数を数える
            totalacc += acc.item()
            loss.backward()
            self.optimizer.step()
        return totalloss, totalacc   

    def evaluation(self, data_loader):
        totalacc = 0
        each_acc = torch.zeros([10,10])
        class_size = torch.zeros([10])
        for index, (images, labels) in enumerate(data_loader):
            images = images.view(-1, 28*28)
            images = images.to(self.device)
            outputs = self.model(images)
            outputs = outputs.to(self.device)
            labels = labels.to(self.device)
            acc = (self.getLabel(outputs)==labels).sum()
            actuallabels = self.getLabel(outputs)
            for p in range(len(labels)):
                    each_acc[labels[p].item()][actuallabels[p].item()] += 1
                    class_size[labels[p].item()] += 1
            totalacc += acc.item()
        return totalacc, each_acc, class_size
    
    def getLabel(self, outputs):
#         print("getLabel() outputs=%s label=%s" %(outputs, outputs.max(1)[1]) )
        return outputs.max(1)[1]
        
if __name__ == "__main__":
    if len(sys.argv) < 2:
        print("Usage python3 NaiveRehearsal [# of rounds][buffer size gain]")
    else:
        MAIN =  main(int(sys.argv[1]), float(sys.argv[2]))
