import sys
import torch.nn as nn
import torch.nn.functional as F
import torchvision.datasets as dsets
import torchvision.transforms as transforms
import torch.optim
import torch.utils.data
import random
from _signal import pause


class MyPerceptron(nn.Module): #多層ネットワークの構造定義
    ## input_size:入力サイズ, hidden_size:中間ユニット数, num_classes:出力ユニット数, device:CPU/GPU ##
    def __init__(self, input_size, hidden_size, num_classes, device): 
        super(MyPerceptron, self).__init__()
        self.fc1 = nn.Linear(input_size, hidden_size) #第1層
        self.fc2 = nn.Linear(hidden_size, hidden_size) #第2層目
        self.fc3 = nn.Linear(hidden_size, num_classes) #第3層目
        self.num_classes = num_classes #出力層のニューロン数
        self.dropout = nn.Dropout(p=0.4) # ドロップアウト40%に設定
    
    def forward(self, x): #出力計算メソッド定義
        out = self.fc1(x) #第1層目
        out = F.relu(out) #relu()
        out = self.dropout(out) # 1層目ドロップアウト
        out = self.fc2(out) #第2層目
        out = F.relu(out) #relu()
        out = self.dropout(out) # 2層目ドロップアウト
        out = self.fc3(out) #第3層
        return out
        
    ##最後の学習結果(パラメータ)およびフィッシャー情報行列を保存##
    def storeParameters(self, diag_fisher_matrix):
        for n, p in self.named_parameters():
            n = n.replace('.', '-')
            self.register_buffer('mean{}'.format(n), p.data.clone())
            self.register_buffer('Fishermatricies{}'.format(n), diag_fisher_matrix[n].clone())
            
    ## 出力ベクトルからラベルを得る ##
    def getLabel(self, outputs):
        return outputs.max(1)[1]
        
class MyMNIST(torch.utils.data.Dataset):
    ## MinistデータセットのうちTargetLabels[]で指定されたラベルのデータだけを集める ##
    def __init__(self, TargetLabels, root, train, transform, download):
        self.transform = transform # データ変換メソッド
        self.download = download # インターネット上からDLするかどうかのフラグ
        self.tran = train # 学習用かそうでないかのフラグ
        ## MINSTデータセットは別途読み込んでおく##
        self.dataset = dsets.MNIST(root='./data', train=True, transform=self.transform, download= \
        self.download)
        self.datasetSize = 0 # データセットのサイズ
        self.indicies = [] # 収集するべきデータのインデックス配列
        self.label = [] # ラベルデータ配列
        ## TargetLabels[]にマッチするレコードの探索 ##
        for p in range(self.dataset.__len__()):
            each_input, each_label = self.dataset.__getitem__(p)
            for n in range(len(TargetLabels)):
                if each_label == TargetLabels[n]:#ラベルが一致する場合
                    self.indicies.append(p) # インデックスをself.indiciesに保持する
                    self.datasetSize += 1 #ラベルの一致するデータの個数を数える
        self.mydata = torch.tensor([self.datasetSize,28,28]) #テンソル型の入力データ配列を用意する
        self.mydata = torch.zeros(self.datasetSize, 28, 28) #上で用意したテンソル型データを初期化
        for p in range(self.datasetSize): #ラベルの一致する入力データセットを作成する
            each_input, each_label = self.dataset.__getitem__(self.indicies[p]) # indicies[]で指定されたレコードを読み込む
            self.mydata[p] = each_input # 読み込んだ入力をmydata[]に積み込む
            self.label.append(each_label) #ラベルデータもself.labelに保持する
        self.data = self.mydata.numpy() # 入力データはTensorからnumpy arrayに変換する必要がある
        
    def __len__(self): #データセットのサイズを返すメソッド
        return self.datasetSize
    
    def __getitem__(self, idx): #各々のデータを返すメソッド
        out_data = self.data[idx]
        out_label =  self.label[idx]
        if self.transform:
            out_data = self.transform(out_data) #ラベルデータの変換
        return out_data, out_label
     
     

class main():     
    def __init__(self, hidden_size, num_division, myLambda, numIterations):
        print(" main() _init_() started!")
        self.input_size = 784
        self.hidden_size = hidden_size
        self.num_classes = 10
        self.batch_size = 8000
        self.num_division = num_division
        self.myLambda = myLambda
        self.label_step = int(10/self.num_division)
        self.criterion = nn.CrossEntropyLoss() #クロスエントロピーロスを使用
        self.totalN = 0 # total number of learned samples
        
        ## テストデータセットを用意 ##
        self.test_dataset = dsets.MNIST(root='./data', train=False, transform=transforms.ToTensor(),download=True)
        test_loader = torch.utils.data.DataLoader(dataset=self.test_dataset, 
                                                  batch_size=10000, shuffle=False)
        
        ## GPUもしくはCPUの選択deviceの自動選択 ##
        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        ## パーセプトロンクラスのインスタンス化 ##
        self.model = MyPerceptron(self.input_size, self.hidden_size, self.num_classes, self.device)
        ## パーセプトロンオブジェクトを演算を実行するデバイスに作成 ##
        self.model = self.model.to(self.device)
        self.optimizer = torch.optim.Adagrad(self.model.parameters(True), lr=0.001) 
        print(self.model) #ネットワーク構造を表示
        print("Lambda=%s" %(self.myLambda))
        print("NumIterations=%s" %(numIterations))
        Acc = [] #正解率
        for n in range(0, self.num_division):
            ##このラウンドで学習させるラベルをtarget_labelsにセットする ##
            target_labels = []
            for l in range(n*self.label_step, (n+1)*self.label_step):
                target_labels.append(l)
            ##target_labels[]に従って学習用データをロードする ##
            self.train_dataset = MyMNIST(target_labels, './data', True, transforms.ToTensor(),True)
            print("Main().datasetsize=%s" %(self.train_dataset.__len__()))
            
            ##tran_datasetからtrain_loaderを用意する ##
            train_loader = torch.utils.data.DataLoader(dataset=self.train_dataset, \
            batch_size=self.batch_size, shuffle=True)

            totalloss = 0 #総合ロスを初期化
            totalacc = 0  #総正解率を初期化
            for i in range(numIterations): # numIteration epoch繰り返す
                if n == 0:
                    err = self.learning(train_loader, 0) # 初期学習
                else:
                    err = self.learning(train_loader, self.myLambda) # 追加学習
                print("%s : err=%s" %(i, err))
                totalloss += err
                train_loss = totalloss / (self.batch_size*len(self.train_dataset)) # 平均誤差算出
                print("averaged train loss = %s" %(train_loss)) #表示
            
            ## 別途バッチサイズを大きくしたtrain_loader を用意する ##
            train_loader = torch.utils.data.DataLoader(dataset=self.train_dataset, batch_size=1, \
            shuffle=True) 
            Fisher = self.getFisherMatrix(train_loader) # fisher行列を計算
            self.model.storeParameters(Fisher) # 学習結果とfisher行列を保存
            
            testerr, AccMatrix, class_size = self.evaluation(test_loader) #テスト用サンプルでの誤差計測
            print("Accuracy matrix = %s" %(AccMatrix))
            for c in range(10):
                print("%s : %s " %(c, (float)(AccMatrix[c][c].item()/class_size[c].item())))
            Acc.append(testerr)
        #誤差の一覧表示
        for n in range(0, len(Acc)):
            print("Acc[%s]=%s" %(n, Acc[n]))

    ## フィッシャー行列の計算 ##
    def getFisherMatrix(self, train_loader):
        self.model.eval() # ドロップアウトは使用しないモードにする
        self.Fisher = {}  # Fisher行列の対角成分
        self.optimizer.zero_grad()
        IsFirstIteration = False
        for index, (images, labels) in enumerate(train_loader):
            images = images.view(-1,28*28)
            batchSize = len(labels)
            images = images.to(self.device) # imageを演算を実行するデバイスに作成
            outputs = self.model(images) # imageに対する出力を求める
            labels = labels.to(self.device) # labelを演算を実行するデバイスに作成
            loss =  self.criterion(outputs, labels).to(self.device) 
            loss.backward()
            
            #現在のタスクに対するフィッシャー情報行列を得る
            for n, p in self.model.named_parameters():
                n = n.replace('.', '-')
                try:
                    ## 最後に保持したフィッシャー情報行列を読み込む ##
                    if index == 0 :
                        self.Fisher[n] =getattr(self.model, \
                        'Fishermatricies{}'.format(n))
                    ## フィッシャー情報行列を作る ##
                    self.Fisher[n] += (1/(index+self.totalN))*((p.grad ** 2)/batchSize - self.Fisher[n])
                except AttributeError:
                    IsFirstIteration = True
                    self.Fisher[n] = (p.grad ** 2) / batchSize
        self.totalN += index # 層学習サンプル数をセット
        return self.Fisher
  
    ## 正則化項の定義 ##
    def EWCPenarty(self):
        loss = 0
        for n, p in self.model.named_parameters():
            n = n.replace('.', '-')
            mean = getattr(self.model, 'mean{}'.format(n)).detach()
            diag_fisher = getattr(self.model, 'Fishermatricies{}'.format(n))
            each_loss = diag_fisher * ((p - mean) ** 2)
            loss += each_loss.sum()
        return loss 
    
    ## 学習 ##
    def learning(self, train_loader, lamda):
        self.model.train() # 学習モード(ドロップアウトを有効にする)
        totalloss = 0 # 変数初期化

        ## train_loaderからチャンク毎にデータを取り出す。##
        for index, (images, labels) in enumerate(train_loader):
            images = images.view(-1,28*28) # imagesは1つのチャンクを表すテンソル。
            images = images.to(self.device)
            ## デバッグ用にラベルを表示（今何を学習しているのかを表示）##
            batchSize = len(labels)
            if index==0:
                print("MyPerceptron.learning() labels=%s" %(labels))
            
            ## 勾配の積算値をゼロリセット ##
            self.optimizer.zero_grad()
            
            outputs = self.model(images) #出力を求める(チャンク分全部を一度に求める)
            labels = labels.to(self.device) #教師データとしてのラベルを演算を実行するデバイスに作成

            if lamda == 0:
                loss = self.criterion(outputs, labels)  # 損失関数=新規サンプルの損失 (初回のみ)
            else:
                loss = self.criterion(outputs, labels) + 0.5 * lamda * self.EWCPenarty() # 損失関数=新規サンプルの損失 + EWCペナルティ
            loss.backward()
            totalloss += loss.item()
            self.optimizer.step() # パラメータ更新
        return totalloss
    
    def evaluation(self, data_loader):
        #self.model.eval() # ドロップアウトは使用しない
        totalacc = 0
        each_acc = torch.zeros([10,10])
        class_size = torch.zeros([10])
        for index, (images, labels) in enumerate(data_loader):
            images = images.view(-1, 28*28)
            images = images.to(self.device)
            outputs = self.model(images)
            labels = labels.to(self.device)
            acc = (self.model.getLabel(outputs)==labels).sum()
            actuallabels = self.model.getLabel(outputs)
            for p in range(len(labels)):
                    each_acc[labels[p].item()][actuallabels[p].item()] += 1
                    class_size[labels[p].item()] += 1
            totalacc += acc.item()
        return totalacc, each_acc, class_size


if __name__ == "__main__":
    if len(sys.argv) < 5:
        print("Usage python3 EWC [# of hiddenUnits][# of rounds][lambda value][number of iterations]")
    else:
        MAIN =  main(int(sys.argv[1]), int(sys.argv[2]), float(sys.argv[3]), int(sys.argv[4]))
